module Main where
    import Text.ParserCombinators.Parsec
    import Directory
    import List

    -- Token type name rules write_commands
    data Token = Token String String [Rule] String deriving (Eq, Show)
    -- Rule pattern init_commands
    data Rule = Rule String String deriving (Eq, Show)

    transformType :: String -> String
    transformType s = tr s
	where tr :: String -> String
	      tr s = 't':'_':tr1 s
	      tr1 :: String -> String
	      tr1 [] = []
	      tr1 (' ':l) = '_':tr1 l
	      tr1 ('*':l) = '_':'p': tr2 l
	      tr1 (c:l) = c:tr1 l
	      tr2 :: String -> String
	      tr2 [] = []
	      tr2 ('*':l) = 'p':tr2 l

    transformAction :: String -> String -> String
    transformAction _ [] = []
    transformAction t ('$':'$':l) = "token.value."++t++transformAction t l
    transformAction t (c:cs) = c:transformAction t cs

    token_types :: [Token] -> String
    token_types ts = (concat.nub.map tt) ts
	where tt :: Token -> String
	      tt (Token "void" _ _ _) = ""
	      tt (Token t _ _ _) = t ++ " " ++ (tr t) ++ ";\n"
	      tr = transformType

    token_case_string :: [Token] -> String
    token_case_string ts = concatMap tcs ts
	where tcs :: Token -> String
	      tcs (Token "char*" n _ _) = "case " ++ n ++ ":\n"
	      tcs _ = ""

    token_lexer :: [Token] -> String
    token_lexer ts = concatMap tl ts
	where tl :: Token -> String
	      tl (Token t n rs _)  = concatMap (rl (transformType t) n) rs
	      rl :: String -> String -> Rule -> String
	      rl t n (Rule p a) = p++" {token.type = "++n++";"++transformAction t a++"return token;}\n"

    token_writer :: [Token] -> String
    token_writer ts = concatMap tw ts
	where tw :: Token -> String
	      tw (Token t n _ a) = "case "++n++": "++transformAction (transformType t) a++"break;\n"

    token_values :: [Token] -> String
    token_values ts = concatMap tv ts
	where tv :: Token -> String
	      tv (Token _ n _ _) = n++",\n"

    token_mapping :: [Token] -> String
    token_mapping ts = concatMap tm ts
	where tm :: Token -> String
	      tm (Token t n _ _) = n++" "++transformType t++" "++t++"\n"

    transformOutFileP :: [Token] -> Parser String
    transformOutFileP ts = do l <- many (
					 try (do {string "#token_types"; return (token_types ts);})
					 <|> try (do {string "#token_case_string"; return (token_case_string ts);})
					 <|> try (do {string "#token_lexer"; return (token_lexer ts);})
					 <|> try (do {string "#token_writer"; return (token_writer ts);})
					 <|> try (do {string "#token_values"; return (token_values ts);})
					 <|> try (do {string "#token_mapping"; return (token_mapping ts);})
					 <|> many1 (noneOf "#")
					 <|> do {c <- char '#'; return "#";}
					)
			      return (concat l)

    transformOutFiles :: [String] -> [Token] -> IO ()
    transformOutFiles os ts = mapM_ (transformOutFile ts) os

    getOutFile :: String -> String
    getOutFile [] = error "internal error"
    getOutFile ('.':'f':'t':'o':[]) = []
    getOutFile (c:cs) = c:getOutFile cs

    transformOutFile :: [Token] -> String -> IO ()
    transformOutFile ts name = do val <- parseFromFile (transformOutFileP ts) name
				  case (val) of
					     Left err -> error ("parse error at " ++ show err)
					     Right x -> writeFile (getOutFile name) x

    concatSepBy :: Parser [a] -> Parser [a] -> Parser [a]
    concatSepBy p1 p2 = do l <- many (p1<|>p2)
			   return (concat l)

    escapeP :: Parser Char
    escapeP = do char '\\'
		 c <- anyChar
		 return c

    patternP :: Parser String
    patternP = do c <- oneOf "/$@%"
		  s <- (many (noneOf ['\\', c] <|> escapeP))
		  char c
		  return s
		  

    ruleP :: Parser Rule
    ruleP = do p <- patternP
	       spacesP
	       c <- commandsP
	       return (Rule p c)

    spaceP :: Parser Char
    spaceP = oneOf " \t\n"

    spacesP :: Parser ()
    spacesP = skipMany spaceP

    ruleSepP :: Char -> Parser ()
    ruleSepP x = try (do spacesP
		         char x
		         spacesP
		     )

    concatP :: Parser [[a]] -> Parser [a]
    concatP p = do l <- p
		   return (concat l)

    commStringP :: Parser String
    commStringP = do c <- between (char '"') (char '"') (concatP (many (do { x <- char '\\'
									   ; y <- anyChar
									   ; return (x:y:[])
									   } <|> many1 (noneOf "\\\""))))
		     return (('"':c)++['"']);

    commandsP :: Parser String
    commandsP = do c <- between (char '{') (char '}') (concatP (many (commStringP <|> commandsP <|> many1 (noneOf "{}\""))))
		   return (('{':c)++"}")

    nameP :: Parser String
    nameP = many1 (alphaNum <|> char '_')

    stringP :: String -> Parser String
    stringP s = do x <- string s
		   y <- many1 spaceP
		   return (x++" ")

    pointerP :: Parser Char
    pointerP = do spacesP
		  p <- char '*'
		  return p

    typeP :: Parser String
    typeP = do m <- (try (stringP "unsigned") <|> try (stringP "signed") <|> try (stringP "long") <|> try (stringP "short"))
	       spacesP
	       t <- typeP
	       return (m++t)
	    <|>	do m <- option "" (try (stringP "struct") <|> try (stringP "union")) 
		   t <- nameP
		   spacesP
		   p <- many pointerP
		   return (concat [m,t,p])

    findTokenP :: Parser [Token]
    findTokenP = do try (string "#token")
		    spaceP
		    spacesP
		    t <- typeP
		    spacesP
		    n <- nameP
		    ruleSepP '='
		    r <- sepBy ruleP (ruleSepP '|')
		    w <- option "" (do ruleSepP '&'
				       commandsP)
		    return [Token t n r w]
		 <|> do char '#'
			return []

    noTokenP :: Parser [a]
    noTokenP = do skipMany1 (noneOf "#")
		  return []

    findTokensP :: Parser [Token]
    findTokensP = concatSepBy findTokenP noTokenP

    findTokens :: String -> IO [Token]
    findTokens name = do val <- parseFromFile findTokensP name
			 case (val) of
				    Left err -> error ("parse error at " ++ show err)
				    Right x -> return x

    getAllFiles :: [String] -> IO [String]
    getAllFiles [] = return []
    getAllFiles (d:ds) = do c <- doesDirectoryExist d
			    if c
			       then do l1 <- getDirectoryContents d
				       getAllFiles ((concatMap (addDir d) l1) ++ ds)
			       else do l1 <- getAllFiles ds
				       return (d:l1)
	where addDir :: String -> String -> [String]
	      addDir d f | head f == '.' = []
			 | True          = [d++('/':f)]

    isCFile :: String -> Bool
    isCFile [] = False
    isCFile ('.':'c':[]) = True
    isCFile ('.':'p':'c':[]) = True
    isCFile (',':'p':'a':'r':'s':'e':'r':[]) = True
    isCFile ('.':'h':[]) = True
    isCFile (_:cs) = isCFile cs

    isOutFile :: String -> Bool
    isOutFile [] = False
    isOutFile ('.':'f':'t':'o':[]) = True
    isOutFile (_:cs) = isOutFile cs

    main = do d <- getCurrentDirectory
	      l <- getAllFiles [d]
	      tokens <- (mapM findTokens (filter isCFile l))
	      transformOutFiles (filter isOutFile l) (concat tokens)
		