module Main where
    import Text.ParserCombinators.Parsec
    import Numeric
    import System

    data Code = Code String String [String] Int Mod Int String [String] deriving (Eq, Show)
    data Mod = M_NOP | M_ZBIT | M_R1 | M_COMBINE deriving (Eq, Show)

    data ZProc = ZProc ZEndian [ZCode] deriving (Eq,Show)
    data ZEndian = Little_Endian | Big_Endian deriving (Eq,Show)
    data ZCode = ZCode String [ZArg] deriving (Eq,Show)
    data ZArg = ZArg String [String] deriving (Eq,Show)

    parseString :: Parser a -> String -> a
    parseString p s = case (parse p "" s) of
					  Left err -> error ("parse error at "++ show err)
					  Right x -> x

    parseTasm :: String -> IO [Code]
    parseTasm f = do val <- parseFromFile codes_p f
		     case (val) of
				Left err -> error ("parse error at "++ show err)
				Right x -> return x

    mod_p :: Parser Mod
    mod_p = do {try (string "NOP"); return M_NOP}
	    <|> do {try (string "ZBIT"); return M_ZBIT}
	    <|> do {try (string "R1"); return M_R1}
	    <|> do {try (string "COMBINE"); return M_COMBINE}

    skip :: Parser a -> Parser ()
    skip p = do p
		return ()

    spaces_p :: Parser ()
    spaces_p = skipMany (oneOf " \t")

    newline_p :: Parser ()
    newline_p = do skip (char '\n') <|> skip (string "\r\n")

    name_p :: Parser String
    name_p = many (alphaNum <|> oneOf "_")

    int_p :: Parser Int
    int_p = do s <- many1 digit
	       (return.fst.head.readDec) s

    args_p :: Parser String
    args_p = do {try (string "\"\""); return ""}
	     <|> many (alphaNum <|> oneOf ",()*'")

    data_p :: Parser [String]
    data_p = many (count 2 hexDigit)

    code_p :: Parser Code
    code_p = do m <- name_p
		spaces_p
		a <- args_p
		spaces_p
		d <- data_p
		spaces_p
		b <- int_p
		spaces_p
		mod <- mod_p
		spaces_p
		c <- int_p
		spaces_p
		s <- name_p
		spaces_p
		o <- data_p
		spaces_p
		newline_p
		return (Code m a d b mod c s o)

    comment_p :: Parser ()
    comment_p = skipMany (skip (do {char '"'; many (noneOf "\""); char '"'})
			  <|> skip (do {try (string "/*"); manyTill anyChar (try (string "*/"))})
			  <|> skipMany1 space
			 )

    codes_p :: Parser [Code]
    codes_p = many (between comment_p comment_p code_p)

    tasm2zasm :: [Code] -> ZProc
    tasm2zasm cs = ZProc Little_Endian (tasm2zasmC cs)

    tasm2zasmC :: [Code] -> [ZCode]
    tasm2zasmC [] = []
    tasm2zasmC cs@((Code name _ _ _ _ _ _ _):_) = (ZCode name r):tasm2zasmC o
	where (r,o) = tasm2zasmArgs name cs

    tasm2zasmArgs :: String -> [Code] -> ([ZArg],[Code])
    tasm2zasmArgs _ [] = ([],[])
    tasm2zasmArgs name (c@(Code n _ _ _ _ _ _ _):cs) | n == name = (tasm2zasmA c:zs,ts)
						     | True      = (zs,c:ts)
						     where (zs,ts) = tasm2zasmArgs name cs

    args :: [(Int,String)] -> [String] -> [String]
    args [] l = l
    args ((0,_):as) os = args as os
    args ((1,a):as) []        = ("#byte "++a):args as []
    args ((1,a):as) ("00":os) = ("#byte "++a):args as os
    args ((1,a):as) (o:os)    = ("#byte ("++a++").|0x"++o):args as os
    args ((2,a):as) []             = ("#word "++a):args as []
    args ((2,a):as) ("00":"00":os) = ("#word "++a):args as os
    args ((2,a):as) (o1:o2:os)     = ("#word ("++a++").|0x"++o1++o2):args as os
    args ((4,a):as) []                       = ("#dword "++a):args as []
    args ((4,a):as) ("00":"00":"00":"00":os) = ("#dword "++a):args as os
    args ((4,a):as) (o1:o2:o3:o4:os)         = ("#dword ("++a++").|0x"++o1++o2++o3++o4):args as os
    args (_:as) os = args as os

    orLastByte :: String -> [String] -> [String]
    orLastByte a (b@(_:_:[]):[]) = ["#byte ("++a++").|0x"++b]
    orLastByte a (b:bs) = b:orLastByte a bs

    tasm2zasmA :: Code -> ZArg
    tasm2zasmA (Code _ pat code bytes M_NOP 1 s orC) 
	= ZArg pat (reverse code++args [(bytes - length code,"@0")] (reverse orC))
    tasm2zasmA (Code _ pat code bytes M_ZBIT 1 s orC)
	= ZArg pat (orLastByte "(@0)<<3" (reverse code++args [(bytes - length code - 1,"@1")] (reverse orC)))
    tasm2zasmA (Code _ pat code bytes M_R1 1 s orC)
	= ZArg pat (reverse code++args [(bytes - length code,"@0-$.end")] (reverse orC))
    tasm2zasmA (Code _ pat code bytes M_COMBINE 1 s orC)
	= ZArg pat (reverse code
		    ++args [((bytes - length code) `div` 2,"@0"),((bytes - length code) `div` 2,"@1")] (reverse orC))

    writeZProc :: String -> ZProc -> IO ()
    writeZProc f p = writeFile f (print_zproc p "")

    print_list_raw :: String -> (a->ShowS) -> [a] -> ShowS
    print_list_raw _ _ [] = id
    print_list_raw _ p (a:[]) = p a
    print_list_raw d p (a:as) = p a . showString d . print_list_raw d p as

    print_list :: String -> String -> String -> (a->ShowS) -> [a] -> ShowS
    print_list o c d p as = showString o . print_list_raw d p as . showString c

    print_l :: (a->ShowS) -> [a] -> ShowS
    print_l = print_list "[" "]\n" " "

    print_zproc :: ZProc -> ShowS
    print_zproc (ZProc e cs) = print_zendian e . print_l print_zcode cs

    print_zendian :: ZEndian -> ShowS
    print_zendian (Little_Endian) = showString "LITTLE ENDIAN\n"
    print_zendian (Big_Endian) = showString "BIG ENDIAN\n"

    print_zcode :: ZCode -> ShowS
    print_zcode (ZCode name as) = showString name . print_l print_zarg as

    print_zarg :: ZArg -> ShowS
    print_zarg (ZArg p ds) = showChar '[' . showString p . showChar ']' . print_l print_zdata ds

    print_zdata :: String -> ShowS
    print_zdata [] = id
    print_zdata d@('#':_) = showString d
    print_zdata d = showChar '`' . showString d . showChar '`'

    main = do (input:output:[]) <- getArgs
	      cs <- parseTasm input
	      writeZProc output (tasm2zasm cs)
