.include <ti83plus.inc>

.export	parser_parse_expr
.export parser_code
.export parser_var_x
.export parser_var_y

;; assumes that only hl should remain intact
PARSER_ADD_FUNC .macro args,function
	.begin_section
	call parser_add_func
	.db args
	.dw end - begin
begin:
	function
end:
	ret
	.end_section
	.endm

PARSER_ADD_PUSH .macro
	call parser_add_push
	.endm
		
;; call a function located at (hl)
call_hl:
	jp (hl)	

;; parser_parse_expr
;;   description:	parse an expression
;;   arguments:		hl:			pointer to the expression  string
;;			bc:			the length of the string
;;   returns:		parser_code:		a function to evaluate the expression, the previous parse is destroyed
;;			flag c:			true on error
;;   registers destroyed:	all
parser_parse_expr:
.begin_section
	ld de,parser_code
	ld (parser_code_ptr),de
	call parser_parse_sum
	ret c
	ld a,b
	or c
	jr nz,not_parsed
	PARSER_ADD_FUNC(0, ret)
	ret
not_parsed:
	scf
	ret
.end_section

;; parser_parse_sum
;;  description:	parse a sum, as part of an expression
;;    arguments:	hl:	pointer to the beginning of the sum expression
;;			bc:	count of bytes still left in the string
;;    returns:		OP1/OP2:	if possible, the result of the expression
;;			flag c:	failed
;;			hl:	the end of the sum expression
;;			bc:	the number of bytes left in the string
;;    registers destroyed:	all
parser_parse_sum:
.begin_section
	ld de,(parser_code_ptr)
	push de
	call parser_parse_prod
	pop de
	ret c
loop:
	;; is there a token left in the string?
	ld a,b
	or c
	jr z,nul

	ld a,(hl)
	cp tAdd
	jr nz,minus
	push bc
	push de
	PARSER_ADD_PUSH
	pop de
	pop bc
	ret c
	
	inc hl
	dec bc

	push de
	call parser_parse_prod
	pop de

	ret c
	push bc
	push de
	PARSER_ADD_FUNC(2,
		B_CALL CAdd
	)
	pop de
	pop bc
	ret c
	jr loop
minus:
	cp tSub
	jr nz,nul
	push bc
	push de
	PARSER_ADD_PUSH
	pop de
	pop bc
	ret c
	
	inc hl
	dec bc
	
	push de
	call parser_parse_prod
	pop de
	ret c

	push bc
	push de
	PARSER_ADD_FUNC(2,
		B_CALL CSub
	)
	pop de
	pop bc
	ret c
	jr loop
nul:
	scf
	ccf
	ret
.end_section
	
;; parser_parse_prod
;;  description:	parse a product, as part of an expression
;;    arguments:	hl:	pointer to the beginning of the product expression
;;			bc:	count of bytes still left in the string
;;    returns:		OP1/OP2:	if possible, the result of the expression
;;			flag c:	failed
;;			hl:	the end of the product expression
;;			bc:	the number of bytes left in the string
;;    registers destroyed:	all
parser_parse_prod:
.begin_section
	ld de,(parser_code_ptr)
	push de
	call parser_parse_elem
	pop de
	ret c
loop:
	;; is there a token left in the string?
	ld a,b
	or c
	jr z,nul

	ld a,(hl)
	cp tMul
	jr nz,div

	push bc
	push de
	PARSER_ADD_PUSH
	pop de
	pop bc
	ret c
	
	inc hl
	dec bc

	push de
	call parser_parse_elem
	pop de
	ret c

	push bc
	push de
	PARSER_ADD_FUNC(2,
		B_CALL CMult
	)
	pop de
	pop bc
	ret c
	jr loop
div:
	cp tDiv
	jr nz,nul
	push bc
	push de
	PARSER_ADD_PUSH
	pop de
	pop bc
	ret c
	
	inc hl
	dec bc
	
	push de
	call parser_parse_elem
	pop de
	ret c

	push bc
	push de
	PARSER_ADD_FUNC(2,
		B_CALL CDiv
	)
	pop de
	pop bc
	ret c
	jr loop
nul:
	scf
	ccf
	ret
.end_section

;; parser_parse_elem
;;  description:	parse an elementary expression
;;    arguments:	hl:	pointer to the beginning of the expression
;;			bc:	count of bytes still left in the string
;;    returns:		OP1/OP2:	if possible, the result of the expression
;;			flag c:	failed
;;			hl:	the end of the expression
;;			bc:	the number of bytes left in the string
;;    registers destroyed:	all
parser_parse_elem:
.begin_section
	ld a,b
	or c
	jr z,end_of_string

	ld a,(hl)
	cp tLParen
	jr nz,x
	inc hl
	dec bc
	call parser_parse_sum

	ld a,b
	or c
	jr z,end_of_string

	ld a,(hl)
	cp tRParen
	inc hl
	dec bc
	scf
	ret nz
	ccf
	ret
x:
	cp tX
	jr nz,y
	ld de,(parser_code_ptr)
	inc hl
	dec bc
	push bc
	PARSER_ADD_FUNC(0,{
		ld hl,parser_var_x
		call parser_load_var
		}
	)
	pop bc
	ret
y:
	cp tY
	jr nz,number
	ld de,(parser_code_ptr)
	inc hl
	dec bc
	push bc
	PARSER_ADD_FUNC(0,{
		ld hl,parser_var_y
		call parser_load_var
		}
	)
	pop bc
	ret
number:
	call parser_parse_number
	ret c
	ld de,(parser_code_ptr)
	push bc
	PARSER_ADD_FUNC(1, {})
	pop bc
	ret

end_of_string:
	scf
	ret
.end_section

;; parser_parse_number
;;   description:	parse a number and put it in OP1/OP2
;;   arguments:		hl:	string
;;			bc:	the length of the string
;;   returns:		OP1/OP2:	a complex number with the value of the string
;;			flag c:		failed
;;			hl:		next token
;;			bc:		length of the string from next token
;;   registers destroyed:	all
#define not_init_token 0x0d
parser_parse_number:
#begin_section
	ld ix,exponent
	ld a,(hl)
	sub t0
	cp 11
	ccf
	ret c
	push hl
	B_CALL COP1Set0
	pop de
	ld hl,OP1M

	push bc
	;; make the mantissa all 0x00
	ld bc,0
	ld a,not_init_token
	ld (OP1M),a
	ld (OP1M+1),bc
	ld (OP1M+3),bc
	ld (OP1M+5),bc
	ld (OP1M+7),bc
	pop bc

	ld (ix+0),0x7F
	ld a,0x0C
	ld (OP1),a

skip_zeros:
	ld a,b
	or c
	jr z,number_end_eos

	ld a,(de)
	inc de
	dec bc
	cp t0
	jr z,skip_zeros

	cp tDecPt
	jr z,skip_zeros2
numeric:
	sub t0
	cp 10
	jr nc,point
	inc (ix+0)
	rld
	or a
	jr z,numeric_end
	inc hl
	ld a,(hl)
	or a
	jr nz,skip_numbers
	ld (hl),not_init_token
numeric_end:
	ld a,b
	or c
	jr z,number_end_eos

	ld a,(de)
	inc de
	dec bc
	jr numeric
	
skip_zeros2:
	ld a,b
	or c
	jr z,number_end_eos
	
	ld a,(de)
	inc de
	dec bc
	dec (ix+0)
	cp t0
	jr z,skip_zeros2
	inc (ix+0)
	jr numeric2
	
point:
	cp tDecPt-t0
	jr nz,number_end
	ld a,b
	or c
	jr z,number_end_eos
	ld a,(de)
	inc de
	dec bc
numeric2:
	sub t0
	cp 10
	jr nc,number_end
	rld
	or a
	jr z,numeric2_end
	inc hl
	ld a,(hl)
	or a
	jr nz,skip_numbers2
	ld (hl),not_init_token
numeric2_end:
	ld a,b
	or c
	jr z,number_end_eos
	ld a,(de)
	inc de
	dec bc
	jr numeric2

number_end:
	dec de
	inc bc
number_end_eos:
	ld a,(hl)
	cp not_init_token
	jr nz,number_end2
	ld (hl),0
	jr number_end3
number_end2:
	cp 0x0c
	jr z,number_end3
	ld a,0
	rld
number_end3:
	ld a,(ix+0)
	ld (OP1+1),a
	scf
	ccf
	ex de,hl
	ret

skip_numbers:
	ld a,b
	or c
	jr z,number_end_eos
	ld a,(de)
	inc de
	dec bc
	inc (ix+0)
	sub t0
	cp 10
	jr c,skip_numbers
	cp tDecPt-t0
	jr nz,number_end
skip_numbers2:
	ld a,b
	or c
	jr z,number_end_eos
	ld a,(de)
	inc de
	dec bc
	sub t0
	cp 10
	jr c,skip_numbers2
	add a,t0
	jr number_end

exponent:
	.db 0x00
#end_section

;; parser_add_push
;;   start parsing the next argument
;;   arguments: none
;;   returns: flag c: failure
;;   registers destroyed:	all except hl
parser_add_push:
#begin_section
	push hl
	ld hl,(parser_code_ptr)
	inc hl
	inc hl
	inc hl
	ld de,parser_code_end
	scf
	ccf
	sbc hl,de
	ccf
	jr c,ret_err

	ld de,(parser_code_ptr)
	ld hl,push_func
	ld bc,3
 	ldir
	ld (parser_code_ptr),de
push_func:
	call PushMCplxO1
	
	scf
	ccf
ret_err:
	pop hl
	ret
#end_section
	
;; parser_add_func
;;   arguments:	just after the calling location:	1 byte:	number of arguments to the function (on FPS and OP1/OP2)
;;							2 bytes:	length of the code
;;		de:	the beginning of the code for this function
;;   returns:	flag c:	failure
;;   registers destroyed:	all except hl
parser_add_func:
#begin_section
	ex (sp),hl
	ld a,(hl)
	or a
	jr z,copy

	push hl
	push de
	ld b,a
	ld a,(OP1)
	ld hl,(FPS)
	ld de,18
	scf
	ccf
args_check_loop:
	sbc hl,de
	and 0x40
	jr nz,copy_pop
	ld a,(hl)
	djnz args_check_loop
	pop de
	pop hl
	inc hl
	inc hl
	inc hl
	push hl
	push de
	call call_hl
	pop de
	pop hl
	dec hl
	ld b,(hl)
	dec hl
	ld c,(hl)
	inc hl
	inc hl
	add hl,bc
	inc hl
	ld (parser_code_ptr),de
	push hl
	;; check if everything fits
	ld h,d
	ld l,e
	ld bc,3+18
	add hl,bc
	ld bc,parser_code_end
	scf
	ccf
	sbc hl,bc		
	ccf			
	jr c,ret_err_nocopy

	;; add a constant
	ld hl,load_const
	ld bc,3
 	ldir
	ld hl,OP1
	ld bc,9
	ldir
	ld hl,OP2
	ld bc,9
	ldir
	ld (parser_code_ptr),de
	
	ld a,(OP1)
	and ~0x40
	ld (OP1),a
	
	scf
	ccf
ret_err_nocopy:
	pop hl
	ex (sp),hl
	ret

load_const:
	call parser_load_const
	
copy_pop:
	pop de
	pop hl
	ld a,(hl)
copy:
	inc hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	inc hl

	;; check if the function fits
	push hl
	ld hl,(parser_code_ptr)
	add hl,bc
	ld de,parser_code_end
	scf
	ccf
	sbc hl,de
	ccf
	jr c,ret_err_copy

	;; copy the function
	call clear_fps
	pop hl
	ld de,(parser_code_ptr)
	ldir
	ld (parser_code_ptr),de
	inc hl
	ex (sp),hl
	scf
	ccf
	ret
	
ret_err_copy:
	call clear_fps
	pop hl
	add hl,bc
	inc hl
	ex (sp),hl
	scf
	ret

clear_fps:
	or a
	jr z,clear_fps_end
	dec a
	jr z,clear_fps_end
	ld b,a
clear_fps_loop:
	push bc
	call PopMCplxO1
	pop bc
	djnz clear_fps_loop
clear_fps_end:
	ld a,(OP1)
	or 0x40
	ld (OP1),a
	ret
#end_section

;; parser_load_const
;;   description:	load a constant to OP1/OP2
parser_load_const:
	ex (sp),hl
	ld de,OP1
	ld bc,9
	ldir
	ld de,OP2
	ld bc,9
	ldir
	ex (sp),hl
	ret

;; parser_load_var
;;   description:	load a variable to OP1/OP2
parser_load_var:
	ld de,OP1
	ld bc,9
	ldir
	ld de,OP2
	ld bc,9
	ldir
	ret

;; variables X and Y
parser_var_x:
	.db [18]0x00
parser_var_y:
	.db [18]0x00

;; pointer to the current end of the code
parser_code_ptr:
	.dw parser_code
	
;; the code comes here
parser_code:
	.db [2048]0x00
parser_code_end:
